!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: IndexManager.gs 25988 2011-06-14 17:55:54Z stever $
!
! Superclass Hierarchy:
!  IndexManager, Object.
!
!=========================================================================

! class created in idxclasses.topaz

removeallmethods IndexManager
removeallclassmethods IndexManager

category: 'For Documentation Installation only'
classmethod: IndexManager
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'IndexManager is provides protocol for managing and maintaining indexed collections.
 The single instance of IndexManager should be accessed via the class message #current.
 In normal use, you will use IndexManager to turn autoCommit for index maintenance
 operations off and on and to define the dirtyObjectCommitThreshold and 
 percentTempObjSpaceCommitThreshold, which are control parameters for autoCommit.

 IndexManager current tracks all instances of UnorderedCollection that have had indexes
 created for them. You may use IndexManager current to #removeAllIndexes, which removes
 indexes from all indexed collections in the system or you may use IndexManager 
 current to #removeAllIncompleteIndexes, which removes partial indexes from collections
 when index creation failed without completing.

 Automatic garbage collection of unreferenced UnorderedCollections is _not_ preformed, so
 one must explicitly remove all indexes from an UnorderedCollection that is no longer used.

 IndexManager may be subclassed.'.
doc documentClassWith: txt.

self description: doc.
%
! ------------------- Class methods for IndexManager
category: 'Accessing'
classmethod: IndexManager
current

  Current == nil ifTrue: [ Current := IndexManager _basicNew initialize ].
  ^ Current
%

category: 'Accessing'
classmethod: IndexManager
autoCommit

^self current autoCommit
%

category: 'Index Maintenance'
classmethod: IndexManager
removeAllIndexes

^ self current removeAllIndexes.
%

category: 'Index Maintenance'
classmethod: IndexManager
removeAllIncompleteIndexes

^ self current removeAllIncompleteIndexes.
%

category: 'Index Maintenance'
classmethod: IndexManager
cleanupDependencyLists

  "Iterate through all dependency lists, cleaning up entries that have removed indexes."

  self current cleanupDependencyLists
%

category: 'Updating'
classmethod: IndexManager
dirtyObjectCommitThreshold: anInteger

self current dirtyObjectCommitThreshold: anInteger
%

category: 'Updating'
classmethod: IndexManager
percentTempObjSpaceCommitThreshold: anInteger

self current percentTempObjSpaceCommitThreshold: anInteger
%

category: 'Updating'
classmethod: IndexManager
autoCommit: aBool

self current autoCommit: aBool
%

category: 'Updating'
classmethod: IndexManager
sessionAutoCommit: aBool

self current sessionAutoCommit: aBool
%

! ------------------- Instance methods for IndexManager

! deleted IndexManager>>addIndex: renamed to IndexManager>>_addIndex: v2.1

category: 'Accessing'
method: IndexManager
_addIndex: anIndex
  "  Add the given index into the collection of all indexes."

  anIndex objectSecurityPolicy: GsIndexingObjectSecurityPolicy.
  self getAllIndexes add: anIndex
%
category: 'Index Maintenance'
method: IndexManager
cleanupDependencyLists

  "Iterate through all dependency lists, cleaning up entries that have removed indexes."
%
category: 'Index Updating'
method: IndexManager
createRcEqualityIndexFor: anNSC on: aPathString withLastElementClass: aClass

  "Replaces UnorderedCollection>>_createEqualityIndexOn:withLastElementClass:commitInterval:, modulo the error handler"


  ( ( aClass canUnderstand: #<= ) or:
  [ aClass == Boolean or: [ aClass == UndefinedObject ] ] )
    ifFalse: [
      ^ { #rtErrBagClassDoesNotSupportRangeOperators . { aClass } }
    ].
^[System disableStoneGemTimeout.
  self _createIndex: self _equalityIndex 
        for: anNSC 
        on: aPathString 
        withLastElementClass: aClass 
        equalityIndexClass:  RcRangeEqualityIndex ]
    ensure: [ System enableStoneGemTimeout ] 
%
category: 'Index Updating'
method: IndexManager
createEqualityIndexFor: anNSC on: aPathString withLastElementClass: aClass

  "Replaces UnorderedCollection>>_createEqualityIndexOn:withLastElementClass:commitInterval:, modulo the error handler"


  ( ( aClass canUnderstand: #<= ) or:
  [ aClass == Boolean or: [ aClass == UndefinedObject ] ] )
    ifFalse: [
      ^ { #rtErrBagClassDoesNotSupportRangeOperators . { aClass } }
    ].
  ^self _createIndex: self _equalityIndex for: anNSC on: aPathString withLastElementClass: aClass
%
category: 'Index Updating'
method: IndexManager
createIdentityIndexFor: anNSC on: aPathString

  "Replaces UnorderedCollection>>_createIdentityIndexOn:commitInterval:"

  ^self _createIndex: self _identityIndex for: anNSC on: aPathString withLastElementClass: nil
%
category: 'Accessing'
method: IndexManager
getAllIndexes

  "Return a collection of all indexes in the system."

  allIndexes == nil 
    ifTrue: [
      allIndexes := RcIdentityBag new.
      allIndexes objectSecurityPolicy: GsIndexingObjectSecurityPolicy.
    ].
  ^allIndexes
%
category: 'Accessing'
method: IndexManager
getAllNSCRoots

  "Return a collection of all nsc's with indexes in the system."

  | set |
  set := IdentitySet new.
  self getAllIndexes do: [:index |
    set add: index nscRoot
  ].
  ^set
%

category: 'Index Maintenance'
method: IndexManager
removeAllIndexes

| sys |
sys := System .
^[
  | roots sz |
  sys setIndexProgressCountTo: UnorderedCollection statValueForRemovingAllIndexes .
  roots := self getAllNSCRoots.
  sz := roots size .
  sys setIndexProgressCountTo: sz.
  1 to: sz do: [:n| | nsc result |
     nsc := roots at: n . 
     result := self removeAllIndexesOn: nsc.
    (nsc == result)
      ifFalse: [ self _error: #rtErrRemoveAllIndexesFailed args: { nsc .  result } ].
    sys decrementProgressCountBy: 1 .
  ].
] ensure:[ sys setIndexProgressCountTo: 0; setIndexProgressCountTo: 0 ] .
%

category: 'Index Maintenance'
method: IndexManager
removeAllIncompleteIndexes

  self getAllNSCRoots do: [:nsc | 
    self removeAllIncompleteIndexesOn: nsc
  ].
%
category: 'Index Updating'
method: IndexManager
removeAllIndexesOn: anNSC

  | result |
result := self removeAllCompleteIndexesOn: anNSC.
result == anNSC
  ifTrue: [
    self removeAllIncompleteIndexesOn: anNSC.
  ].
^result
%
category: 'Index Updating'
method: IndexManager
removeAllCompleteIndexesOn: anNSC
  "Remove all complete indexes for anNSC. If all of the anNSC's indexes can be removed, 
   this method returns the receiver.
   If an error occurs during index removal, it may not be possible to 
   commit the current transaction later.
  "
  | iList rootIndexes hasNonRootIndex |
  (iList := anNSC _indexedPaths) ifNil: [ ^ anNSC ].

  hasNonRootIndex := false.
  rootIndexes := { } .
  " scan each entry to see if a non-root index is present "
  1 to: iList size by: 2 do: [ :i |
    " if it is a root index ... "
    (((iList at: i + 1) == 1) and: [ (iList at: i) isComplete])
        " add it to a list "
        ifTrue: [ rootIndexes add: (iList at: i) ]
        " otherwise flag that we're really not removing ALL indexes "
        ifFalse: [ hasNonRootIndex := true ]
  ].
  ^ [  | result |
	  IndexManager current executeStartingIndexMaintenance: [
		result := anNSC _removeAllRootIndexes: rootIndexes 
                		hasNonRootIndex: hasNonRootIndex.
		result == anNSC ifTrue: [
		  rootIndexes do: [:each | self _removeIndex: each.  ].
		].
      ].
      result
	] onSynchronous: Error do:[ :ex | | errNumPreventingCommit |
		" check for recursive signal "
		errNumPreventingCommit := ErrorSymbols at: #rtErrPreventingCommit .
		ex number == errNumPreventingCommit ifTrue: [
			IndexManager current autoCommit ifTrue: [ self removeIncompleteIndex ].
			ex pass
		] ifFalse: [
			ex resignalAs:( Exception _new: errNumPreventingCommit args: { ex } )
		].
	]
%

category: 'Index Updating'
method: IndexManager
removeAllIncompleteIndexesOn: anNSC

  "Remove all incomplete indexes for anNSC. 
   If all of the anNSC's indexes can be removed, this method returns the receiver."

  | iList rootIndexes |
  (iList := anNSC _indexedPaths) ifNil: [ ^ anNSC ].

  rootIndexes := { } .
  " scan each entry to see if a non-root index is present "
  1 to: iList size by: 2 do: [ :i |
    " if it is an incomplete root index ... "
    (((iList at: i + 1) == 1) and: [ (iList at: i) isComplete not])
        " add it to a list "
        ifTrue: [ rootIndexes add: (iList at: i) ]
  ].

  IndexManager current executeStartingIndexMaintenance: [
        rootIndexes do: [:indexObj |
           indexObj nscRoot _undoIndexCreation: indexObj 
				pathTerm: indexObj _findFirstUnsharedPathTerm.
          self _removeIndex: indexObj.
    ].
  ].
  ^ rootIndexes size
%

category: 'Index Updating'
method: IndexManager
removeEqualityIndexFor: anNSC on: aPathString

  self _removeIndex: self _equalityIndex for: anNSC on: aPathString
%
category: 'Index Updating'
method: IndexManager
removeIdentityIndexFor: anNSC on: aPathString

  self _removeIndex: self _identityIndex for: anNSC on: aPathString
%

! deleted IndexManager>>removeIndex: renamed to IndexManager>>_removeIndex: v2.1

category: 'Accessing'
method: IndexManager
_removeIndex: anIndex

  "Remove the given index from the collection of all indexes.
   Make sure that IVs are niled in anIndex, since it will
   survive (in rc redo log) until abort or commit."

  self getAllIndexes remove: anIndex ifAbsent: [ nil ].
  anIndex _clear.
%
category: 'Accessing'
method: IndexManager
resetAllIndexes

  allIndexes := nil.
%
category: 'Index Updating - Private'
method: IndexManager
_createIndex: indexType for: anNSC on: aPathString withLastElementClass: aClass

^[System disableStoneGemTimeout.
  self _createIndex: indexType 
        for: anNSC on: aPathString 
        withLastElementClass: aClass 
        equalityIndexClass:  RangeEqualityIndex ]
    ensure: [ System enableStoneGemTimeout ]
%

category: 'Index Updating - Private'
method: IndexManager
_createIndex: indexType for: anNSC on: aPathString withLastElementClass: aClass 
		equalityIndexClass:  aRangeEqualityIndex


" see if incomplete index for same path "
anNSC _hasIncompleteIndexes ifTrue: [ ^self _error: #rtErrCollectionWithIncompleteIndex].
[ 
  | pathArray index indexes  lastConstraint original | 
  pathArray := aPathString asArrayOfPathTerms.
  " check if an identity index already exists for the path "
  indexes := anNSC  _findIndexesWithPath: pathArray.
  indexes do: [ :idxObj |
    (indexType == self _identityIndex) ifTrue: [
        idxObj isIdentityIndex ifTrue: [ ^ nil ]
	] ifFalse: [
        idxObj isRangeEqualityIndex ifTrue: [ ^ nil ]
	].
  ].

  "No index already exists for collection, lets create one"
  (indexType == self _identityIndex) ifTrue: [
      lastConstraint := anNSC getLastElementConstraintOnPath: aPathString.
      ( lastConstraint ~~ nil and: [ anNSC _isRangeable: lastConstraint ] )
        ifTrue: [
          1 to: indexes size do: [ :i |
            (indexes at: i) isRangeEqualityIndex ifTrue: [ ^ nil ]
          ].
          index := aRangeEqualityIndex newWithLastElementClass: lastConstraint
        ]
        ifFalse: [ index := IdentityIndex new ].
    ] ifFalse: [
        index := aRangeEqualityIndex newWithLastElementClass: aClass.
    ].

  " don't need index dictionary if path length is one "
  ( pathArray size == 1 and: [ index isRangeEqualityIndex ] ) ifFalse: [
      | indexDict |
      anNSC _calculateIndexDictionarySize: pathArray size.
      indexDict := anNSC _getIndexDictionary.
      index indexDictionary: indexDict.
      "afford an opportunity to resize dictionary before adding new elements"
      indexDict rebuildTable: anNSC _getIndexDictionaryCreationSize for: index.
  ].
  index nscRoot: anNSC.

  anNSC _getIndexList buildPathTermsFor: index with: pathArray.
  anNSC _putInWriteSet.

  self _addIndex: index.
  original := index preIndexCreation.

  " optimize creation of indexes on elements of the NSC itself "
  index isIndexOnRootNsc ifTrue: [
      index addDirectMappingsFor: anNSC _asIdentityBag indexList: anNSC _indexedPaths
    ] ifFalse: [
      self _addMappingsFor: anNSC on: index
    ].

  index postIndexCreation: original.
] onSynchronous: Error do:[ :ex | | errNumPreventingCommit |
	errNumPreventingCommit := ErrorSymbols at: #rtErrPreventingCommit .
    ex number == errNumPreventingCommit ifTrue:[
       " remove this exception and resignal "
       self autoCommit ifTrue: [ anNSC removeIncompleteIndex ].
       ex pass
    ] ifFalse:[
       ex resignalAs:( Exception _new: errNumPreventingCommit args: { ex } )
    ]
].
^ nil
%

category: 'Private'
method: IndexManager
_doCommit
  | systm |
  systm := System .
  systm commitTransaction ifFalse:[
     self _errorCouldNotCommitDuringIndexCreation
  ].
  systm transactionMode == #manualBegin ifTrue:[
     systm beginTransaction
  ].
%
category: 'Private'
method: IndexManager
_equalityIndex

  ^ #equality
%
category: 'Private'
method: IndexManager
_identityIndex

  ^ #identity
%
category: 'Index Updating - Private'
method: IndexManager
_removeIndex: indexType for: anNSC on: aPathString

  | iList indexes index pathArray |
  anNSC _checkIndexPathExpression: aPathString.

  iList := anNSC _indexedPaths.
  iList ifNil: [ self _error: #rtErrNoIndexForPath args: { aPathString } ].

  pathArray := aPathString asArrayOfPathTerms.
  " check if an index exists for the path string "
  indexes := anNSC _findIndexesWithPath: pathArray.

  indexes isEmpty ifFalse: [
      " make sure it is an identity index or equality index"
      indexes do: [ :iObj |
        (indexType == self _identityIndex)
          ifTrue: [
            ( iObj isIdentityIndex or:
              [ anNSC _isRangeable: iObj lastElementClass ] )
                ifTrue: [ index := iObj ]
          ]
          ifFalse: [
            iObj isRangeEqualityIndex ifTrue: [ index := iObj ]
          ]
      ].
    ].

  " no index was found with the given path "
  index ifNil: [ ^ self _error: #rtErrNoIndexForPath args: { aPathString }  ].
  [
    IndexManager current executeStartingIndexMaintenance: [
      index preIndexRemoval.
      anNSC _removeIndex: index .
      self _removeIndex: index .
    ].
  ] onSynchronous: Error do:[ :ex | | errNumPreventingCommit |
	  " check for recursive signal "
	  errNumPreventingCommit := ErrorSymbols at: #rtErrPreventingCommit .
	  ex number == errNumPreventingCommit ifTrue: [
		  IndexManager current autoCommit ifTrue: [ self removeIncompleteIndex ].
		  ex pass
	  ] ifFalse: [
		  ex resignalAs:( Exception _new: errNumPreventingCommit args: { ex } )
	  ].
  ]

%

category: 'Testing'
method: IndexManager
shouldCommit
"Answer true if a commit should be performed during index maintenance.
   Note - that this method should only be executed from within a
   executeStartingIndexMaintenance: block, since _autoCommitPolicy
   may not have been initialized correctly otherwise."

| policy |
self autoCommit ifFalse: [ ^ false ].

policy := self _autoCommitPolicy .
(policy == nil or: [ "bug34941"policy == true]) ifTrue: [ ^ false ].

^ policy shouldCommit
%

category: 'Testing'
method: IndexManager
_autoCommitPolicy

^ (System _sessionStateAt: 11)
%

category: 'Testing'
method: IndexManager
_clearAutoCommitPolicy

System _sessionStateAt: 11 put: nil
%

category: 'Testing'
method: IndexManager
_initializeAutoCommitPolicy

System _sessionStateAt: 11 put: (IndexManagerAutoCommitPolicy on: self)
%

category: 'Initializing'
method: IndexManager
initialize

  "When creating initial instance, make sure all IVs are set, since we don't
   want to dirty the object during normal operation."

  self autoCommit: false .       
  self getAllIndexes.
  self useKeyEncryption: true.
  self dirtyObjectCommitThreshold: SmallInteger maximumValue . "disabled by default"
  "was 75% ,  Gs64 v3.0 change to 60 for new AlmostOutOfMemory handling. "
  self percentTempObjSpaceCommitThreshold: 60.
  self objectSecurityPolicy: GsIndexingObjectSecurityPolicy.
%

category: 'Accessing'
method: IndexManager
autoCommit
  "Use session-specific value if sessionAutoCommit has been set (bug40602)"

| sac |
(sac := self sessionAutoCommit) ~~ nil
  ifTrue: [ ^sac ].
autoCommit == nil ifTrue:[ autoCommit := false ].
^autoCommit
%

category: 'Accessing'
method: IndexManager
useKeyEncryption
  
  ^ useKeyEncryption
%

category: 'Updating'
method: IndexManager
useKeyEncryption: aBool
  
  useKeyEncryption := aBool

%

category: 'Updating'
method: IndexManager
autoCommit: aBool

autoCommit := aBool
%

category: 'Accessing'
method: IndexManager
sessionAutoCommit

^SessionTemps current at: #INDEX_MANAGER_AUTO_COMMIT ifAbsent: []
%

category: 'Updating'
method: IndexManager
sessionAutoCommit: aBoolOrNil
"Override persistent autoCommit on a session by session basis.
 If <aBoolOrNil> is nil, remove override."

aBoolOrNil == nil
  ifTrue: [SessionTemps current removeKey: #INDEX_MANAGER_AUTO_COMMIT ifAbsent: []]
  ifFalse: [ SessionTemps current at: #INDEX_MANAGER_AUTO_COMMIT put: aBoolOrNil ]
%

category: 'Accessing'
method: IndexManager
dirtyObjectCommitThreshold

^ dirtyObjectCommitThreshold
%

category: 'Accessing'
method: IndexManager
rcBtreeLeafNodeClass

  ^ RcBtreeLeafNode
%

category: 'Accessing'
method: IndexManager
rcBtreeBasicLeafNodeClass

  self useKeyEncryption ifTrue: [ ^ RcBtreeBasicLeafNode ].

  ^ RcBtreeLeafNode
%

category: 'Accessing'
method: IndexManager
btreeBasicLeafNodeClass

  self useKeyEncryption ifTrue: [ ^ BtreeBasicLeafNode ].

  ^ BtreeLeafNode
%

category: 'Accessing'
method: IndexManager
btreeLeafNodeClass

  ^ BtreeLeafNode
%

category: 'Accessing'
method: IndexManager
sortNodeClass

  self useKeyEncryption ifTrue: [ ^ BasicSortNode ].

  ^ SortNode
%

category: 'Updating'
method: IndexManager
dirtyObjectCommitThreshold: anInteger

dirtyObjectCommitThreshold := anInteger
%

category: 'Accessing'
method: IndexManager
percentTempObjSpaceCommitThreshold

^ percentTempObjSpaceCommitThreshold
%

category: 'Updating'
method: IndexManager
percentTempObjSpaceCommitThreshold: anInteger

  anInteger _isSmallInteger ifFalse:[ anInteger _validateClass: SmallInteger ].
  (anInteger < 10 or:[ anInteger > 100]) ifTrue:[
     OutOfRange new name:'commit threshold' min: 10 max: 100 actual: anInteger;
       details: 'invalid commit threshold' ;  signal
  ].
  percentTempObjSpaceCommitThreshold := anInteger
%

category: 'Updating'
method: IndexManager
objectSecurityPolicy: anObjectSecurityPolicy

"Assigns the receiver and the allIndexes collection to anObjectSecurityPolicy."

<primitive: 2001>
| prot |
prot := System _protectedMode .
[
  anObjectSecurityPolicy == GsIndexingObjectSecurityPolicy
    ifFalse: [ self _error: #objectSecurityPolicyNotSharedDepListObjectSecurityPolicy ].

  self _objectSecurityPolicy: anObjectSecurityPolicy.
  self getAllIndexes objectSecurityPolicy: anObjectSecurityPolicy.
] ensure:[
  prot _leaveProtectedMode
]
%

category: 'Index Maintenance'
method: IndexManager
commitIndexMaintenance: indexObj at: progressCount

self shouldCommit ifTrue: [
    indexObj progress: progressCount.
    self _doCommit .
].
%

category: 'Private'
method: IndexManager
execStartingIndexMaintenance: aBlock
"Behavior of index maintenance with respect to transaction state:

When in manual mode:
    When IndexManager current autoCommit is true:
        When originally outside of a transaction:
            Begin a transaction.
            Commit the final transaction of index creation.
            End outside of a transaction.
        When originally inside of a transaction:
            Commit the final transaction of index creation.
            End outside of a transaction.
When in auto mode:
    When IndexManager current autoCommit is true:
        Commit the final transaction of index creation.
"


  | inAutoCommit autoCommitPolicy |

(inAutoCommit := self autoCommit) ifTrue: [ | systm |
    systm := System .
    systm transactionMode == #manualBegin 
      ifTrue:
        [systm inTransaction 
           ifFalse:
             [systm needsCommit
                ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .
              systm beginTransaction]].
    self _initializeAutoCommitPolicy
].

autoCommitPolicy := self _autoCommitPolicy.
(autoCommitPolicy == nil or: [ "bug34941"autoCommitPolicy == true]) ifTrue: [
    aBlock value.
  ] ifFalse: [
    [ 
 		autoCommitPolicy evaluate: aBlock for: self 
    ] ensure: [ 
        inAutoCommit ifTrue: [ self _clearAutoCommitPolicy ].
	].
  ].

inAutoCommit ifTrue: [
  System commitTransaction ifFalse:[ self _errorCouldNotCommitDuringIndexCreation].
].
%

category: 'Index Maintenance'
method: IndexManager
executeStartingIndexMaintenance: aBlock
"Disable STN_GEM_TIMEOUT, then perform index maintenance.

Behavior of index maintenance with respect to transaction state:

When in manual mode:
    When IndexManager current autoCommit is true:
        When originally outside of a transaction:
            Begin a transaction.
            Commit the final transaction of index creation.
            End outside of a transaction.
        When originally inside of a transaction:
            Commit the final transaction of index creation.
            End outside of a transaction.
When in auto mode:
    When IndexManager current autoCommit is true:
        Commit the final transaction of index creation."
| sys |
sys := System .
^ [ 
    sys disableStoneGemTimeout .
    self execStartingIndexMaintenance: aBlock
  ] ensure:[ sys enableStoneGemTimeout ] .
%


category: 'Updating Indexes - Private'
method: IndexManager
_addMappingsFor: anNSC
on: indexObj

"Adds the mappings in the index dictionary, dependency lists, and so on for all
 elements in the receiver."

| pathTerm object bag iList |

" since this index may use path terms from other indexes,
find the first one that is not shared by another index (only
need to add mappings from that point onward) "
iList := anNSC _indexedPaths.

pathTerm := indexObj _findFirstUnsharedPathTerm.
pathTerm ifNil: [ ^ anNSC " index is subsumed by an existing index " ].

bag := anNSC _asIdentityBag.
self autoCommit
  ifTrue: [ anNSC _lockForIndexCreation ].

" see if need to traverse to the first unshared path term "
indexObj firstPathTerm == pathTerm
  ifTrue: [
	1 to: bag size do: [ :i |
	  object := bag _at: i.
	  pathTerm addMappingsForObject: object logging: false.
	  self commitIndexMaintenance: indexObj at: i.
	]
  ]
  ifFalse: [
	1 to: bag size do: [ :i |
	  object := bag _at: i.
	  " traverse up to the point of the unshared path term "
	  object := indexObj traverse: object upTo: pathTerm.
	  pathTerm addMappingsForObject: object logging: false.
	  self commitIndexMaintenance: indexObj at: i.
	]
  ].
^ anNSC
%

category: 'Querying'
method: IndexManager
nscsWithBadIndexes
"
	IndexManager current nscsWithBadIndexes.
"
	^ self getAllNSCRoots reject: [:each | each auditIndexes = 'Indexes are OK'].
%

category: 'Querying'
method: IndexManager
nscsWithIncompleteIndexes
"
	IndexManager current nscsWithIncompleteIndexes.
"
	^ self getAllNSCRoots select: [:each | each _hasIncompleteIndexes].
%

category: 'Querying'
method: IndexManager
usageReport
"
	Returns a string describing all indexes, the collection sizes, and 
	the (visible) methods that could create these indexes.

	IndexManager current usageReport.
"

	| indexes methods keys stream |
	(stream := (WriteStream on: String new))
		lf; nextPutAll: 'Index Type - path - lastElementClass';
		lf; tab; nextPutAll: 'Collection size (collection OOP)';
		lf; tab; tab; nextPutAll: 'Index creation methods';
		lf.
	indexes := self _reportDataForIndexes.
	methods := self _reportDataForMethods.
	keys := (indexes collect: [:each | each key]) asSet asSortedCollection.
	keys do: [:eachKey | 
		stream lf; nextPutAll: eachKey.
		(indexes select: [:each | each key = eachKey]) do: [:each | 
			stream lf; tab; nextPutAll: each value.
		].
		(methods select: [:each | each key = eachKey]) do: [:each | 
			stream lf; tab; tab; nextPutAll: each value.
		].
		stream lf.
	].
	^stream contents.
%

category: 'Querying'
method: IndexManager
_reportDataForIndexes
		"private"
	| list |
	list := OrderedCollection new.
	self getAllIndexes do: [:each | | array firstStr string |
		string := each class name copyFrom: 1 to: each class name size - 5.
		(5 < string size and: [(string copyFrom: 1 to: 5) = 'Range']) ifTrue:[
           string := string copyFrom: 6 to: string size
        ].
        firstStr := string .
		string := each inject: '' into: [:string :pathTerm | string , '.' , pathTerm name].
		string := string copyFrom: 2 to: string size.
        list add: { firstStr . 
                    string  .
			   (each isKindOf: RangeEqualityIndex) ifTrue: [each lastElementClass name] ifFalse: [''].
			    each nscRoot .
			    each isComplete } .
	].
	list := list collect: [:each | | nsc |
		nsc := each at: 4.
		Association new
			key: ((each at: 1) , '-' , (each at: 2) , 
					((each at: 3) notEmpty ifTrue: ['-' , (each at: 3)] ifFalse: ['']))
			value: (nsc size printString , ' (' , nsc asOop printString , ')' , 
					   ((each at: 5) ifTrue: [''] ifFalse: [' - INCOMPLETE'])).
	].
	^ list
%

category: 'Querying'
method: IndexManager
_reportDataForMethods
		"private"

	| indexCreationSelectors classOrganizer smallList bigList |
	classOrganizer := ClassOrganizer new.
	smallList := OrderedCollection new.
	indexCreationSelectors := #(
		#'createEqualityIndexOn:'
		#'createEqualityIndexOn:withLastElementClass:'
		#'createEqualityIndexOn:withLastElementClass:commitInterval:'
		#'createIdentityIndexOn:'
		#'createRcEqualityIndexOn:'
		#'createRcEqualityIndexOn:withLastElementClass:'
	).
	indexCreationSelectors do: [:each | | arrays |
		arrays := classOrganizer sendersOf: each.
		1 to: arrays first size do: [:i | 
			smallList add: (Array
				with: (arrays first at: i)
				with: each
				with: (arrays last at: i)).
		].
	].
	smallList := smallList reject: [:each | each first inClass == UnorderedCollection].
	smallList := smallList reject: [:each | each first inClass name == #'IndexedQueryExtensionsTestCase'].
	bigList := OrderedCollection new.
	smallList do: [:each | 
		| selector pieces firstPiece lastPiece source index |
		selector := each at: 2.
		pieces := selector subStrings: $:.
		pieces := pieces copyFrom: 1 to: pieces size - 1.
		firstPiece := pieces first.
		lastPiece := pieces last.
		source := each first sourceString.
		index := (each at: 3) - 1.
		[
			0 < (index := source indexOfSubCollection: firstPiece startingAt: index + 1).
		] whileTrue: [
			| readStream array word |
			readStream := ReadStream on: (source copyFrom: index to: source size).
			array := { ((word := readStream nextWord) copyFrom: 7 to: word size - 8) .
						nil . nil . nil . nil } .
			word := readStream nextWord.
			word := word copyFrom: 2 to: word size - 1.
			[''';.' includes: word last] whileTrue: [word := word copyFrom: 1 to: word size - 1].
			array at: 2 put: word.
			2 <= pieces size ifTrue: [
				word := readStream nextWord; nextWord.
				word := word reject: [:char | ''';.' includes: char].
			] ifFalse: [
				word := ''.
			].
			array 
				at: 3 put: word;
				at: 4 put: each first inClass name;
				at: 5 put: each first selector  .
			bigList add: array.
		].
	].
	bigList := bigList collect: [:each | 
		Association new
			key: ((each at: 1) , '-' , (each at: 2) , 
					((each at: 3) notEmpty ifTrue: ['-' , (each at: 3)] ifFalse: [''])) 
			value: ((each at: 4) , '>>#' , (each at: 5)).
	].
	^ bigList
%

! make sure default values are initialized
run
IndexManager current .
^ true
%
