!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: fractio2.gs 24812 2010-12-07 18:27:47Z lalmarod $
!
! Superclass Hierarchy:
!   Fraction, Number, Magnitude, Object.
!
!=========================================================================

category: 'For Documentation Installation only'
classmethod: Fraction
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'A Fraction is a Number represented as the ratio of two Integers.'.
doc documentClassWith: txt.

txt := (GsDocText new) details:
'The numerator of the Fraction (an Integer).'.
doc documentInstVar: #numerator with: txt.

txt := (GsDocText new) details:
'The denominator of the Fraction (an Integer).'.
doc documentInstVar: #denominator with: txt.

txt := (GsDocText new) details:
'Do not send class Fraction the message new.  Fractions created in that way are
 meaningless and cannot be handled properly by GemStone''s associative access
 mechanism.  To create a new Fraction, use one of the instance creation methods
 listed here.'.
doc documentClassCategory: #'Instance Creation' with: txt.

self description: doc.
%

category: 'Instance Creation'
classmethod: Fraction
fromStream: aStream

"Returns a Fraction from the stream.  The stream must contain two Integers
 separated by a slash.  (There may be blanks around the slash.)  Generates an
 error if the stream contains anything else, or if an attempt is made to read
 beyond the end of the stream."

| ch n d |

self _checkReadStream: aStream forClass: CharacterCollection.

ch := aStream next.
[ ch isEquivalent: $ ]
whileTrue:
  [ ch := aStream next ].
aStream skip: -1.
n := Integer fromStream: aStream.
ch := aStream next.
[ ch isEquivalent: $ ]
whileTrue:
  [ ch := aStream next ].
aStream skip: -1.
(aStream next isEquivalent: $/)
ifFalse:[ ^ self _errIncorrectFormat: aStream ].
ch := aStream next.
[ ch isEquivalent: $ ]
whileTrue:
  [ ch := aStream next ].
aStream skip: -1.
d := Integer fromStream: aStream.
^ Fraction numerator: n denominator: d
%

category: 'Instance Creation'
classmethod: Fraction
numerator: numInt denominator: denomInt

"Returns an instance of Fraction with numerator numInt and denominator
 denomInt.  If that Fraction can be reduced, this method returns the
 corresponding Integer instead.  The result is made invariant.

 If either argument (numerator or denominator) is not an Integer, that
 argument is truncated to the corresponding Integer."

^ (self _basicNew _setNumerator: numInt denominator: denomInt) 
    _reduce immediateInvariant
%

category: 'Private'
method: Fraction
_setNumerator: n denominator: d

"Truncates n and d before assigning to numerator and denominator,
 and moves sign to numerator."

| denom numer |
denom := d truncated abs.
numer := n truncated.
(d < 0) ifTrue: [ numer := numer negated] .
^ self _numerator: numer denominator: denom
%

category: 'Private'
classmethod: Fraction
_numerator: n denominator: d
  ^ (self _basicNew _numerator: n denominator: d) _reduce immediateInvariant
%
category: 'Private'
classmethod: Fraction
_noreduce_numerator: n denominator: d
  "caller responsible for immediateInvariant"
  ^ (self _basicNew _numerator: n denominator: d) 
%

category: 'Private'
method: Fraction
_numerator: n denominator: d
  "caller responsible for _reduce if desired, and immediateInvariant"
  n _isInteger ifFalse:[ n _validateClass: Integer ].
  d _isInteger ifFalse:[ d _validateClass: Integer ].
  (d == 0) ifTrue: [ ^ n _errorDivideByZero ].
  numerator := n .
  denominator := d .
   ^ self
%


category: 'Private'
method: Fraction
_reduce

"reduce to form using least common denominator."
| gcd numer denom |
( numer := numerator) == 0 ifTrue: [ ^0 ].
gcd := numer gcd: (denom := denominator) .
numer := numer // gcd.
denom := denom // gcd.
(denom == 1) ifTrue:[ 
  ^ numer 
  ].
numerator := numer .
denominator := denom .
^ self 
%

category: 'Accessing'
method: Fraction
denominator

"Returns the denominator of the receiver."

^denominator
%

method: Fraction
numerator

"Returns the numerator of the receiver."

^numerator
%

method: Fraction
instVarAt: anIndex put: aValue

"Disallowed.  You may not change the value of a Fraction."

self shouldNotImplement: #instVarAt:put:
%

category: 'Formatting'
method: Fraction
asString

"Returns a String of the form 'numerator/denominator'."

| result |
result := numerator asString .
result add: $/ ; addAll: denominator asString .
^ result
%

category: 'Converting'
method: Fraction
asScaledDecimal: scale

"Returns a ScaledDecimal representation of the receiver."

^ ScaledDecimal numerator: numerator denominator: denominator scale: scale.
%

! asFloat rewritten for fix 36173
category: 'Converting'
method: Fraction
asFloat

"Returns an instance of SmallDouble or Float that has the value of the receiver."
| num denom nParts dParts mantiss exp |

(num := numerator) ifNil:[ ^ PlusSignalingNaN ] .
(denom := denominator) ifNil:[ ^ PlusSignalingNaN ] .

"avoid loss of precision by handling mantissa and exponents separately"
nParts := num _floatParts .
dParts := denom _floatParts .
mantiss := (nParts at:1) / (dParts at:1) .
exp := (nParts at:2) - (dParts at:2) .
^ mantiss * (2.0 raisedToInteger: exp ) 
%


category: 'Converting'
method: Fraction
asDecimalFloat

"Returns an instance of DecimalFloat that has the value of the receiver."
| num denom |
(num := numerator) ifNil:[ ^ PlusSignalingNaN ] .
(denom := denominator) ifNil:[ ^ PlusSignalingNaN ] .
^ num asDecimalFloat / denom asDecimalFloat
%

category: 'Converting'
method: Fraction
asFraction

"(R) Returns the receiver."

^self
% 

category: 'Converting'
method: Fraction
asFixedPoint: scale

^ FixedPoint numerator: numerator denominator: denominator scale: scale
%

category: 'Converting'
method: Fraction
_coerce: aNumber

"Reimplemented from Number."
| f |
f := aNumber asFraction .
f _isInteger ifTrue:[ 
  ^ Fraction _noreduce_numerator: f denominator: 1 
].
^ f
%

category: 'Converting'
method: Fraction
_generality

"Reimplemented from Number."

^70
%

category: 'Accessing'
method: Fraction
_getKind

| num denom |
(num := numerator) ifNil:[ ^ 5 "nan"].
(denom := denominator) ifNil:[ ^ 5 ].
denom == 0 ifTrue:[ ^ 3 "infinity" ].
num == 0 ifTrue:[ ^ 4 "zero"].
^ 1 "normal"
%

method: Fraction
sign
"Returns 1 if the receiver is greater than zero, -1 if the receiver is less
 than zero, and zero if the receiver is zero."

  ^ numerator sign
%

category: 'Comparing'
method: Fraction
<= aNumber

"Returns true if the receiver is less than or equal to aNumber; returns false
 otherwise."

aNumber _isNumber ifTrue:[ 
  (aNumber isKindOf: Fraction) ifTrue: [ | otherNum num |
    (otherNum := aNumber numerator) ifNil:[ ^ false ].
    (num := numerator) ifNil:[ ^ false ].
    otherNum == 0 ifTrue: [ 
      ^ num <= 0
    ].
    ^ (num * aNumber denominator) <=  (denominator * otherNum)
  ].
].
^self _retry: #<= coercing: aNumber
%

category: 'Comparing'
method: Fraction
< aNumber

"Returns true if the receiver is less than aNumber; returns false otherwise."

aNumber _isNumber ifTrue:[ 
  (aNumber isKindOf: Fraction) ifTrue: [ | otherNum num |
    (otherNum := aNumber numerator) ifNil:[ ^ false ].
    (num := numerator) ifNil:[ ^ false ].
    otherNum == 0 ifTrue: [  
      ^ num < 0
    ].
    ^ (num * aNumber denominator) < (denominator * otherNum)
  ].
].
^self _retry: #< coercing: aNumber
%

category: 'Comparing'
method: Fraction
= aNumber

"Returns true if the receiver is equal to aNumber; returns false otherwise."
aNumber _isNumber ifFalse:[ ^ false ].
(aNumber isKindOf: Fraction ) ifTrue: [ | otherNum num |
  (otherNum := aNumber numerator) ifNil:[ ^ false ].
  (num := numerator) ifNil:[ ^ false ].
  otherNum == 0 ifTrue: [
    ^ num == 0
  ].
  ^ otherNum = num and:[ aNumber denominator = denominator ]
].
^ self _retry: #= coercing: aNumber
%

category: 'Comparing'
method: Fraction
> aNumber

aNumber _isNumber ifTrue:[ 
  (aNumber isKindOf: Fraction) ifTrue: [ | otherNum num |
    (otherNum := aNumber numerator) ifNil:[ ^ false ].
    (num := numerator) ifNil:[ ^ false ].
    otherNum == 0 ifTrue: [  
      ^ num > 0
    ].
    ^ (num * aNumber denominator) > (denominator * otherNum)
  ].
].
^self _retry: #> coercing: aNumber
%

category: 'Comparing'
method: Fraction
>= aNumber

aNumber _isNumber ifTrue:[ 
  (aNumber isKindOf: Fraction) ifTrue: [ | otherNum num |
    (otherNum := aNumber numerator) ifNil:[ ^ false ].
    (num := numerator) ifNil:[ ^ false ].
    otherNum == 0 ifTrue: [  
      ^ num >= 0
    ].
    ^ (num * aNumber denominator) >= (denominator * otherNum)
  ].
].
^self _retry: #>= coercing: aNumber
%

! hash inherited from Number

category: 'Truncation and Rounding'
method: Fraction
truncated

"Returns the integer that is closest to the receiver, on the same side
 of the receiver as zero is located."

^numerator quo: denominator
%

method: Fraction
roundTo: aNumber

"Returns the multiple of aNumber that is nearest in value to the receiver."
| r |
aNumber = 0 ifTrue:[ ^ 0 ] .
r := (self / aNumber) rounded * aNumber .
aNumber _isFloat ifTrue:[ ^ r asFloat ].
^ r asFraction 
%

method: Fraction
roundedHalfToEven

"Returns the integer nearest in value to the receiver. If the receiver is 
exactly halfway between two integers, return the even one."

| remainder |
remainder := (numerator abs \\ denominator).
^remainder * 2 = denominator
	ifFalse: [self rounded]
	ifTrue: [| truncated | truncated := self truncated.
			truncated even 
				ifTrue: [truncated]
				ifFalse: [truncated + self sign]]
%

method: Fraction
truncateTo: aNumber

"Returns the multiple of aNumber that is closest to the receiver, on
 the same side of the receiver as zero is located.  In particular,
 returns the receiver if the receiver is a multiple of aNumber."

| t |
aNumber = 0 ifTrue:[ ^ 0 ].
t := (self quo: aNumber) * aNumber .
aNumber _isFloat ifTrue:[ ^ t asFloat ].
^ t asFraction
%


category: 'Arithmetic'
method: Fraction
* aNumber

"Returns the result of multiplying the receiver by aNumber."

(aNumber isKindOf: Fraction) ifTrue: [
  ^ self class _numerator: (numerator * aNumber numerator)
           denominator: (denominator * aNumber denominator) 
] ifFalse: [
  (aNumber _isInteger and:[ aNumber = denominator ]) ifTrue:[
     ^ numerator
  ].
  ^ self _retry: #* coercing: aNumber
]
%


category: 'Arithmetic'
method: Fraction
+ aFraction

"Returns the sum of the receiver and aFraction."

(aFraction isKindOf: Fraction) ifTrue:[ 
  | myDenom otherDenom commonDenominator newNumerator 
    gcd myDemon_div_gcd |
  (myDenom := denominator) = (otherDenom := aFraction denominator) ifTrue: [
    | myNumer otherNumer |
    (myNumer := numerator) = (otherNumer := aFraction numerator) ifTrue:[
      myDenom _bottomBit == 0 ifTrue:[
        ^ self class _numerator: myNumer denominator: (myDenom bitShift: -1)
      ]
    ].
    ^ self class _numerator: (myNumer + otherNumer ) denominator: myDenom 
  ].
  gcd := myDenom gcd: otherDenom .
  myDemon_div_gcd := myDenom // gcd .
  commonDenominator := myDemon_div_gcd * otherDenom .
  newNumerator := (numerator * (otherDenom // gcd ))
                 + (aFraction numerator *  myDemon_div_gcd) .
  ^ self class _numerator: newNumerator denominator: commonDenominator
] ifFalse: [
  ^self _retry: #+ coercing: aFraction
]
%

category: 'Arithmetic'
method: Fraction
- aFraction

"Returns the difference between the receiver and aFraction."
(aFraction isKindOf: Fraction) ifTrue: [
  ^ self + (Fraction _noreduce_numerator:  0 - aFraction numerator 
		 denominator: aFraction denominator) 
] ifFalse: [
  ^ self _retry: #- coercing: aFraction
]
%

category: 'Arithmetic'
method: Fraction
/ aNumber

"Returns the result of dividing the receiver by aFraction."

aNumber _isInteger ifTrue:[
  aNumber = 0 ifTrue: [ ^ self _errorDivideByZero ].
  ^ self class numerator: numerator denominator: (denominator * aNumber)
].
(aNumber isKindOf: Fraction) ifTrue:[
  ^ self * aNumber _reciprocal
].
^ self _retry: #/ coercing: aNumber
%

category: 'Arithmetic'
method: Fraction
negated

"Returns a Number that is the negation of the receiver."

^ (self class _noreduce_numerator:  0 - numerator denominator: denominator) 
	immediateInvariant
%

category: 'Private'
method: Fraction
_reciprocal

"Returns 1 divided by the receiver.  Generates an error if the receiver is 0.
 Result is not reduced."
| numer denom |
((numer := numerator) == 0 or:[ numer == nil]) ifTrue: [ ^ self _errorDivideByZero].
denom := denominator .
numer < 0 ifTrue:[ denom := 0 - denom . numer := 0 - numer  ].

^ ( self class _noreduce_numerator: denom denominator: numer) 
%

category: 'Arithmetic'
method: Fraction
reciprocal

"Returns 1 divided by the receiver.  Generates an error if the receiver is 0.
 Result is reduced."
| numer denom |
((numer := numerator) == 0 or:[ numer == nil]) ifTrue: [ ^ self _errorDivideByZero].
denom := denominator .
(numer == 1) ifTrue:[ ^ denom ].
numer < 0 ifTrue:[ 
  (numer == -1) ifTrue:[ ^ 0 - denom ].
  denom := 0 - denom . numer := 0 - numer .
].
^  self class _numerator: denom denominator: numer 
%

category: 'Storing and Loading'
classmethod: Fraction
loadFrom: passiveObj

"Reads from passiveObj the passive form of an object.  Converts the object to
 its active form by loading the information into a new instance of the receiver.
 Returns the new instance."

| inst num den |

passiveObj readNamedIV.
num := passiveObj ivValue.
passiveObj readNamedIV.
den := passiveObj ivValue.
  
passiveObj skipNamedInstVars.

inst := self numerator: num denominator: den.
passiveObj hasRead: inst.
^inst.
%

category: 'Storing and Loading'
method: Fraction
writeTo: passiveObj

"Converts the receiver to its passive form and writes that information on
 passiveObj."

"Reimplemented from Number since the receiver has a non-literal representation."

^super basicWriteTo: passiveObj
%

category: 'Testing'
method: Fraction
even

"(R) Returns true if the receiver is an even integer, false otherwise."

 denominator = 1 ifFalse: [ ^ false ].
 ^ numerator even

%

category: 'Testing'
method: Fraction
odd

" (R) Returns true if the receiver is an odd integer, false otherwise."

 denominator = 1 ifFalse: [ ^ false ].
 ^ numerator odd
%

category: 'Indexing Support'
method: Fraction
_isNaN

"Returns whether the receiver is quiet NaN or signaling NaN.
 This method is only to be used by the indexing subsystem."

^ (numerator == nil) or: [denominator == nil]
%
category: 'Testing'
method: Fraction
isZero

"Returns true if the receiver is zero."

^ numerator = 0 .
%

