!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: positioPortable.gs 25549 2011-03-24 21:51:57Z jfoster $
!
! Based on Pharo-1.2.1 implementation
!
! Superclass Hierarchy:
!   PositionableStream, Stream, Object.
!
!=========================================================================

! class created by installStreamHierarchy.topaz

removeallmethods PositionableStream
removeallclassmethods PositionableStream

category: 'For Documentation Installation only'
classmethod: PositionableStream
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'ANSI compliant implementation of PositionableStream is an abstract 
 superclass that provides additional protocol appropriate to Streams 
 whose objects are externally named by indices. Concrete subclasses 
 are ReadStream and WriteStream.'.
doc documentClassWith: txt
.

txt := (GsDocText new) details:
'A SequenceableCollection; the sequence of objects that the receiver may
 access.'.
doc documentInstVar: #collection with: txt.

txt := (GsDocText new) details:
'A SmallInteger; the current position reference for accessing the collection.'.
doc documentInstVar: #position with: txt.

txt := (GsDocText new) details:
'A SmallInteger; the position in the collection of the last readable character.'.
doc documentInstVar: #readLimit with: txt.

self description: doc.
%

category: 'Portable Methods'
classmethod: PositionableStream
isLegacyStreamImplementation

^false
%

category: 'Portable Methods'
classmethod: PositionableStream
isPortableStreamImplementation

^true
%


category: 'Instance Creation'
classmethod: PositionableStream
on: aCollection from: firstIndex to: lastIndex 
"Answer an instance of me, streaming over the elements of aCollection 
starting with the element at firstIndex and ending with the one at 
lastIndex."

	^self basicNew on: (aCollection copyFrom: firstIndex to: lastIndex)
%

category: 'Instance Creation'
classmethod: PositionableStream
on: aCollection
"Answer an instance of me, streaming over the elements of aCollection."

^self basicNew on: aCollection.
%

category: 'Accessing'
method: PositionableStream
contents
"Answer with a copy of my collection from 1 to readLimit."

^collection copyFrom: 1 to: readLimit
%

category: 'Accessing'
method: PositionableStream
peek
"Answer what would be returned if the message next were sent to the 
receiver. If the receiver is at the end, answer nil."

| nextObject |
self atEnd ifTrue: [^nil].
nextObject := self next.
position := position - 1.
^nextObject
%

category: 'Accessing'
method: PositionableStream
peek2
"Peeks at the second incoming object."

| nextObject |
position + 1 >= readLimit ifTrue: [^nil].
nextObject := self next; next.
position := position - 2.
^nextObject
%

category: 'Accessing'
method: PositionableStream
skip: anInteger 
"Set the receiver's position to be the current position+anInteger. A 
subclass might choose to be more helpful and select the minimum of the 
receiver's size and position+anInteger, or the maximum of 0 and 
position+anInteger for the repositioning."

self position: position + anInteger
%

category: 'Accessing'
method: PositionableStream
skipTo: anObject 
"Set the access position of the receiver to be past the next occurrence of 
anObject. Answer whether anObject is found."

[self atEnd]
	whileFalse: [self next = anObject ifTrue: [^true]].
^false
%

category: 'Testing'
method: PositionableStream
atEnd
"Returns true if the receiver cannot access any more objects, false if it can."

^position >= readLimit
%

category: 'Testing'
method: PositionableStream
isEmpty
"Answer whether the receiver's contents has no elements."

"Returns true if both the set of past and future sequence values of
the receiver are empty. Otherwise returns false"

^ self atEnd and: [position = 0]
%

category: 'Positioning'
method: PositionableStream
position
"Answer the current position of accessing the sequence of objects."

^position
%

category: 'Positioning'
method: PositionableStream
position: anInteger
"Set the current position for accessing the objects to be anInteger, as long 
as anInteger is within the bounds of the receiver's contents. If it is not, 
create an error notification."

(anInteger >= 0 and: [anInteger <= readLimit])
	ifTrue: [position := anInteger]
	ifFalse: [self positionError: anInteger]
%

category: 'Positioning'
method: PositionableStream
positionError: anInteger
"Returns an error message that anInteger is out of bounds of the Collection."

^ self _error: #rtErrBadStreamPosition args: { anInteger }
%

category: 'Positioning'
method: PositionableStream
reset
"Set the receiver's position to the beginning of the sequence of objects."

position := 0
%

category: 'Positioning'
method: PositionableStream
setToEnd
"Set the position of the receiver to the end of the sequence of objects."

position := readLimit
%

category: 'Positioning'
method: PositionableStream
on: aCollection

collection := aCollection.
readLimit := aCollection size.
position := 0.
self reset
%

category: 'Accessing'
method: PositionableStream
skipSeparators
[self atEnd]
	whileFalse:
	[self next isSeparator ifFalse: [^ self position: self position-1]]
%

method: PositionableStream
upToAll: aCollection 
"Answer a subcollection from the current access position to the occurrence (if any, but not inclusive) of aCollection. If aCollection is not in the stream, answer the entire rest of the stream."

| startPos endMatch result |
startPos := self position.
(self match: aCollection) 
	ifTrue: [endMatch := self position.
		self position: startPos.
		result := self next: endMatch - startPos - aCollection size.
		self position: endMatch.
		^ result]
	ifFalse: [self position: startPos.
		^ self upToEnd] 
%

category: 'Accessing'
method: PositionableStream
upToEnd
"Answer a subcollection from the current access position through the last element of the receiver."

| newStream |
newStream := (collection species new: 100) writeStream.
[self atEnd] whileFalse: [ newStream nextPut: self next ].
^ newStream contents
%

category: 'Accessing'
method: PositionableStream
_collection

"Returns the collection of the receiver."

^collection
%

category: 'Accessing'
method: PositionableStream
next: amount
"Answer the next amount elements of my collection. Must override 
because default uses self contents species, which might involve a large 
collection."

| newArray |
amount < 0 ifTrue: [ ^self error: 'amount may not be less than 0' ].
newArray := collection species new: amount.
1 to: amount do: [:index | newArray at: index put: self next].
^newArray
%

category: 'Accessing'
method: PositionableStream
next: n into: aCollection
"Read n objects into the given collection.
Return aCollection or a partial copy if less than
n elements have been read."
^self next: n into: aCollection startingAt: 1
%

category: 'Accessing'
method: PositionableStream
next: n into: aCollection startingAt: startIndex
"Read n objects into the given collection. 
Return aCollection or a partial copy if less than
n elements have been read."
| obj |
0 to: n-1 do:[:i|
	(obj := self next) == nil ifTrue:[^aCollection copyFrom: 1 to: startIndex+i-1].
	aCollection at: startIndex+i put: obj].
^aCollection
%

category: 'Accessing'
method: PositionableStream
nextLine
"Answer next line (may be empty) without line end delimiters, or nil if at end.
Let the stream positioned after the line delimiter(s).
Handle a zoo of line delimiters CR, LF, or CR-LF pair"

self atEnd ifTrue: [^nil].
^self upToAnyOf: {Character cr. Character lf} do: [:char | char = Character cr ifTrue: [self peekFor: Character lf]]
%

category: 'Accessing'
method: PositionableStream
peekFor: anObject
"Answer false and do not move over the next element if it is not equal to 
the argument, anObject, or if the receiver is at the end. Answer true 
and increment the position for accessing elements, if the next element is 
equal to anObject."

| nextObject |
self atEnd ifTrue: [^false].
nextObject := self next.
"peek for matching element"
anObject = nextObject ifTrue: [^true].
"gobble it if found"
position := position - 1.
^false
%

category: 'Adding'
method: PositionableStream
next: anInteger putAll: aCollection startingAt: startIndex
"Store the next anInteger elements from the given collection."
(startIndex = 1 and:[anInteger = aCollection size])
	ifTrue:[^self nextPutAll: aCollection].
^self nextPutAll: (aCollection copyFrom: startIndex to: startIndex+anInteger-1)
%

category: 'Accessing'
method: PositionableStream
upToAnyOf: subcollection do: aBlock
"Answer a subcollection from the current access position to the occurrence (if any, but not inclusive) of any object in the collection.
Evaluate aBlock with this occurence as argument.
If no matching object is found, don't evaluate aBlock and answer the entire rest of the receiver."

| stream ch |
stream := WriteStream on: collection species new.	
[ self atEnd or: [ (subcollection includes: (ch := self next)) and: [aBlock value: ch. true] ] ] 
	whileFalse: [ stream nextPut: ch ].
^stream _collection "private stream, no need to copy"
%

category: 'Positioning'
method: PositionableStream
match: subCollection 
"Set the access position of the receiver to be past the next occurrence of the subCollection. Answer whether subCollection is found.  No wildcards, and case does matter."
| pattern startMatch |
pattern := ReadStream on: subCollection.
startMatch := nil.
[ pattern atEnd ] whileFalse: 
	[ self atEnd ifTrue: [ ^ false ].
	self next = pattern next 
		ifTrue: [ pattern position = 1 ifTrue: [ startMatch := self position ] ]
		ifFalse: 
			[ pattern position: 0.
			startMatch ifNotNil: 
				[ self position: startMatch.
				startMatch := nil ] ] ].
^ true
%

category: 'Accessing'
method: PositionableStream
upTo: anObject 
"Answer a subcollection from the current access position to the 
occurrence (if any, but not inclusive) of anObject in the receiver. If 
anObject is not in the collection, answer the entire rest of the receiver."
| newStream element |
newStream := WriteStream on: collection species new.
[self atEnd or: [(element := self next) = anObject]]
	whileFalse: [newStream nextPut: element].
^newStream contents
%

category: 'Accessing'
method: PositionableStream
upToAny: aCollection
"Answer a subcollection from the current access position to the 
occurrence (if any, but not inclusive) of any objects in the given collection in the receiver. If 
any of these is not in the collection, answer the entire rest of the receiver."
	
| newStream element |
newStream := WriteStream on: collection species new.
[self atEnd or: [aCollection includes: (element := self next)]]
	whileFalse: [newStream nextPut: element].
^newStream contents
%

category: 'Accessing'
method: PositionableStream
upToEnd
"Answer a subcollection from the current access position through the last element of the receiver."

| newStream |
newStream := WriteStream on: collection species new.
[self atEnd] whileFalse: [ newStream nextPut: self next ].
^ newStream contents
%

category: 'Testing'
method: PositionableStream
atBeginning
"Answer true if the stream is positioned at the beginning"

^position == 0
%
